// Vibe coded with AI
(globalThis as any).canvas_2017 = function (canvas: HTMLCanvasElement) {
  const isStandalone = canvas.getAttribute("data-standalone") === "true";
  // Configuration interface for the checkerboard effect
  interface CheckerboardConfig {
    fps: number; // frames per second
    color1: string; // first checkerboard color
    color2: string; // second checkerboard color
    opacity: number; // opacity of each checkerboard (0-1)
    speedX1: number; // horizontal speed of first checkerboard (pixels per second)
    speedY1: number; // vertical speed of first checkerboard (pixels per second)
    speedX2: number; // horizontal speed of second checkerboard (pixels per second)
    speedY2: number; // vertical speed of second checkerboard (pixels per second)
    baseTileSize: number; // base size of checkerboard tiles
    sizeVariation: number; // maximum variation in tile size (pixels)
    sineFrequency1: number; // frequency of first sine wave for size variation
    sineFrequency2: number; // frequency of second sine wave for size variation
    sineOffset: number; // offset between the two sine waves (radians)
    rotation: number; // rotation in degrees for the entire pattern
    rotation2: number; // rotation in degrees for the entire pattern
  }

  // Default configuration
  const config: CheckerboardConfig = {
    fps: 30,
    color1: "#1A1C17",
    color2: "#1A1C17",
    opacity: 0.3,
    speedX1: -0.02, // moving left slowly
    speedY1: -0.01, // moving up slowly
    speedX2: -0.015, // moving left (slightly slower)
    speedY2: 0.012, // moving down slowly
    baseTileSize: 200,
    sizeVariation: 1.5,
    sineFrequency1: 0.0005,
    sineFrequency2: 0.0008,
    sineOffset: Math.PI / 2, // 90 degrees offset
    rotation: 2, // 5 degree rotation
    rotation2: -2, // 5 degree rotation
  };

  // Get the canvas context
  const ctx = canvas.getContext("2d");
  if (!ctx) {
    console.error("Could not get canvas context");
    return () => {};
  }

  // Make canvas transparent
  if (isStandalone) {
    canvas.style.backgroundColor = "#737D60";
  } else {
    canvas.style.backgroundColor = "transparent";
  }

  // Variables to track position and animation
  let width = canvas.width;
  let height = canvas.height;
  let animationFrameId: number;
  let lastFrameTime = 0;
  const frameInterval = 1000 / config.fps;

  // Position offsets for the two checkerboards (centered)
  let offset1X = 0;
  let offset1Y = 0;
  let offset2X = 0;
  let offset2Y = 0;

  // Time variable for sine wave calculation
  let time = 0;

  // Convert rotation to radians
  const rotationRad = (config.rotation * Math.PI) / 180;
  const rotationRad2 = (config.rotation2 * Math.PI) / 180;

  // Update canvas dimensions when resized
  const updateDimensions = () => {
    width = canvas.width = canvas.clientWidth;
    height = canvas.height = canvas.clientHeight;
  };

  // Calculate the diagonal length of the canvas (to ensure rotation covers corners)
  const calculateDiagonal = () => {
    return Math.sqrt(width * width + height * height);
  };

  // Draw a single checkerboard pattern scaled from center with rotation
  const drawCheckerboard = (
    offsetX: number,
    offsetY: number,
    tileSize: number,
    color1: string,
    color2: string,
    opacity: number,
    rotationRad: number,
  ) => {
    ctx.globalAlpha = opacity;

    // Get the center of the viewport
    const centerX = width / 2;
    const centerY = height / 2;

    // Save the current transformation state
    ctx.save();

    // Move to the center of the canvas, rotate, then move back
    ctx.translate(centerX, centerY);
    ctx.rotate(rotationRad);

    // Calculate the number of tiles needed to cover the rotated canvas
    // We need to use the diagonal length to ensure we cover the corners when rotated
    const diagonal = calculateDiagonal();
    const tilesX = Math.ceil(diagonal / tileSize) + 6; // Added extra tiles for rotation
    const tilesY = Math.ceil(diagonal / tileSize) + 6;

    // Calculate how many tiles fit from center to edge (in each direction)
    const halfTilesX = Math.ceil(tilesX / 2);
    const halfTilesY = Math.ceil(tilesY / 2);

    // Adjust the offset to be relative to the center
    // The modulo ensures the pattern repeats smoothly even with scaling
    const adjustedOffsetX = offsetX % (tileSize * 2);
    const adjustedOffsetY = offsetY % (tileSize * 2);

    // Draw the checker pattern, centered on the viewport
    for (let y = -halfTilesY; y <= halfTilesY; y++) {
      for (let x = -halfTilesX; x <= halfTilesX; x++) {
        // Determine if this tile should be colored (creating checker pattern)
        // We add a large number to ensure (x+y) is always positive for the modulo
        if ((x + y + 1000) % 2 === 0) {
          ctx.fillStyle = color1;
        } else {
          ctx.fillStyle = color2;
        }

        // Calculate the position of this tile relative to the center
        // The adjusted offset creates the movement effect
        const posX = (x * tileSize) + adjustedOffsetX;
        const posY = (y * tileSize) + adjustedOffsetY;

        // Draw the tile
        ctx.fillRect(
          posX - tileSize / 2,
          posY - tileSize / 2,
          tileSize,
          tileSize,
        );
      }
    }

    // Restore the transformation state
    ctx.restore();
  };

  // Animation loop
  const animate = (currentTime: number) => {
    animationFrameId = requestAnimationFrame(animate);

    // Control frame rate
    if (currentTime - lastFrameTime < frameInterval) {
      return;
    }

    // Calculate the time elapsed since the last frame
    const dt = currentTime - lastFrameTime;
    lastFrameTime = currentTime;

    // Increment time for sine wave calculation
    time += dt;

    // Update the position offsets based on speed and elapsed time
    offset1X += config.speedX1 * dt;
    offset1Y += config.speedY1 * dt;
    offset2X += config.speedX2 * dt;
    offset2Y += config.speedY2 * dt;

    // Calculate the tile sizes using sine waves
    const tileSize1 = config.baseTileSize +
      Math.sin(time * config.sineFrequency1) * config.sizeVariation;
    const tileSize2 = config.baseTileSize +
      Math.sin(time * config.sineFrequency2 + config.sineOffset) *
        config.sizeVariation;

    // Clear canvas
    ctx.clearRect(0, 0, width, height);

    // Draw the two checkerboards
    drawCheckerboard(
      offset1X,
      offset1Y,
      tileSize1,
      config.color1,
      "transparent",
      config.opacity,
      rotationRad,
    );

    drawCheckerboard(
      offset2X,
      offset2Y,
      tileSize2,
      config.color2,
      "transparent",
      config.opacity,
      rotationRad2,
    );

    // Reset global alpha
    ctx.globalAlpha = 1.0;
  };

  // Initialize the animation
  const init = () => {
    // Set up resize handler
    window.addEventListener("resize", updateDimensions);

    // Initial setup
    updateDimensions();

    // Start animation
    lastFrameTime = performance.now();
    animationFrameId = requestAnimationFrame(animate);
  };

  // Start the animation
  init();

  // Return cleanup function
  return () => {
    window.removeEventListener("resize", updateDimensions);
    cancelAnimationFrame(animationFrameId);
  };
};
